from laktory import models


def test_catalog():
    m = models.resources.databricks.Catalog(lookup_existing={"name": "0"})


def test_cluster():
    m = models.resources.databricks.Cluster(lookup_existing={"cluster_id": "0"})


def test_dbfsfile():
    m = models.resources.databricks.DbfsFile(lookup_existing={"path": "0"})


def test_directory():
    m = models.resources.databricks.Directory(lookup_existing={"path": "0"})


def test_group():
    m = models.resources.databricks.Group(lookup_existing={"display_name": "0"})


def test_job():
    m = models.resources.databricks.Job(lookup_existing={"id": "0"})


def test_metastore():
    m = models.resources.databricks.Metastore(lookup_existing={"metastore_id": "0"})


def test_notebook():
    m = models.resources.databricks.Notebook(lookup_existing={"path": "0"})


def test_serviceprincipal():
    m = models.resources.databricks.ServicePrincipal(
        lookup_existing={"application_id": "0"}
    )


def test_table():
    m = models.resources.databricks.Table(lookup_existing={"name": "0"})


def test_user():
    m = models.resources.databricks.User(lookup_existing={"user_id": "0"})


def test_warehouse():
    m = models.resources.databricks.Warehouse(lookup_existing={"id": "0"})


if __name__ == "__main__":
    test_catalog()
    test_cluster()
    test_dbfsfile()
    test_directory()
    test_group()
    test_job()
    test_metastore()
    test_notebook()
    test_serviceprincipal()
    test_table()
    test_user()
    test_warehouse()



import uuid
from typing import Any
from typing import Callable
from typing import Literal
from typing import Union
from pydantic import model_validator

from laktory.constants import DEFAULT_DFTYPE
from laktory.models.basemodel import BaseModel
from laktory.models.datasources import DataSourcesUnion
from laktory.models.datasources import BaseDataSource
from laktory.models.datasinks import DataSinksUnion
from laktory.models.pipelinenodeexpectation import PipelineNodeExpectation
from laktory.models.transformers.basechain import BaseChain
from laktory.models.transformers.sparkchain import SparkChain
from laktory.models.transformers.sparkchainnode import SparkChainNode
from laktory.models.transformers.polarschain import PolarsChain
from laktory.models.transformers.polarschainnode import PolarsChainNode
from laktory.spark import SparkSession
from laktory.types import AnyDataFrame
from laktory._logger import get_logger

logger = get_logger(__name__)


class PipelineNode(BaseModel):
    """
    Pipeline base component generating a DataFrame by reading a data source and
    applying a transformer (chain of dataframe transformations). Optional
    output to a data sink. Some basic transformations are also natively
    supported.

    Attributes
    ----------
    add_layer_columns:
        If `True` and `layer` not `None` layer-specific columns like timestamps
        are added to the resulting DataFrame.
    dlt_template:
        Specify which template (notebook) to use if pipeline is run with
        Databricks Delta Live Tables. If `None` default laktory template
        notebook is used.
    drop_duplicates:
        If `True`:
            - drop duplicated rows using `primary_key` if defined or all
              columns if not defined.
        If list of strings:
            - drop duplicated rows using `drop_duplicates` as the subset.
    drop_source_columns:
        If `True`, drop columns from the source after read and only keep
        columns resulting from executing the SparkChain.
    expectations:
        List of expectations for the DataFrame. Can be used as warnings, drop
        invalid records or fail a pipeline.
    layer:
        Layer in the medallion architecture
    name:
        Name given to the node. Required to reference a node in a data source.
    primary_key:
        Name of the column storing a unique identifier for each row. It is used
        by the node to drop duplicated rows.
    source:
        Definition of the data source
    sink:
        Definition of the data sink
    transformer:
        Spark or Polars chain defining the data transformations applied to the
        data source.
    timestamp_key:
        Name of the column storing a timestamp associated with each row. It is
        used as the default column by the builder when creating watermarks.


    Examples
    --------
    A node reading stock prices data from a CSV file and writing a DataFrame
    to disk as a parquet file.
    ```py
    import io
    from laktory import models

    node_yaml = '''
        name: brz_stock_prices
        layer: BRONZE
        source:
            format: CSV
            path: ./raw/brz_stock_prices.csv
        sink:
            format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/brz_stock_prices
    '''

    node = models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute(spark)
    ```

    A node reading stock prices from an upstream node and writing a DataFrame
    to a data table.
    ```py
    import io
    from laktory import models

    node_yaml = '''
        name: slv_stock_prices
        layer: SILVER
        source:
          node_name: brz_stock_prices
        sink:
          catalog_name: hive_metastore
          schema_name: default
          table_name: slv_stock_prices
        transformer:
          nodes:
          - with_column:
              name: created_at
              type: timestamp
              sql_expr: data.created_at
          - with_column:
              name: symbol
              sql_expr: data.symbol
          - with_column:
              name: close
              type: double
              sql_expr: data.close
    '''

    node = models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute(spark)
    ```
    """

    add_layer_columns: bool = True
    dlt_template: Union[str, None] = "DEFAULT"
    dataframe_type: Literal["SPARK", "POLARS"] = DEFAULT_DFTYPE
    description: str = None
    drop_duplicates: Union[bool, list[str], None] = None
    drop_source_columns: Union[bool, None] = None
    transformer: Union[SparkChain, PolarsChain, None] = None
    expectations: list[PipelineNodeExpectation] = []
    layer: Literal["BRONZE", "SILVER", "GOLD"] = None
    name: Union[str, None] = None
    primary_key: str = None
    sink: Union[DataSinksUnion, None] = None
    source: DataSourcesUnion
    timestamp_key: str = None
    _output_df: Any = None
    _parent: "Pipeline" = None
    _source_columns: list[str] = []

    @model_validator(mode="before")
    @classmethod
    def push_dftype_before(cls, data: Any) -> Any:
        dftype = data.get("dataframe_type", None)
        if dftype:
            if "source" in data.keys():
                if isinstance(data["source"], dict):
                    data["source"]["dataframe_type"] = data["source"].get(
                        "dataframe_type", dftype
                    )
            if "transformer" in data.keys():
                if isinstance(data["transformer"], dict):
                    data["transformer"]["dataframe_type"] = data["transformer"].get(
                        "dataframe_type", dftype
                    )

        return data

    @model_validator(mode="after")
    def push_dftype_after(self) -> Any:
        dftype = self.user_dftype
        if dftype:
            self.source.dataframe_type = self.source.user_dftype or dftype
            if self.transformer:
                self.transformer.dataframe_type = self.transformer.user_dftype or dftype
                self.transformer.push_dftype()
        return self

    @model_validator(mode="after")
    def default_values(self) -> Any:
        # Default values
        if self.layer == "BRONZE":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.layer == "SILVER":
            if self.drop_source_columns is None:
                self.drop_source_columns = True
            if self.drop_duplicates is not None and self.primary_key:
                self.drop_duplicates = True

        #
        # if self.layer == "GOLD":
        #     if self.drop_source_columns is None:
        #         self.drop_source_columns = False
        #     if self.drop_duplicates is not None:
        #         self.drop_duplicates = False
        #

        # Generate node id
        if self.name is None:
            self.name = str(uuid.uuid4())

        return self

    @model_validator(mode="after")
    def update_children(self) -> Any:

        # Assign node to sources
        for s in self.get_sources():
            s._parent = self

        # Assign node to sinks
        if self.sink:
            self.sink._parent = self

        # Assign node to transformers
        if self.transformer:
            self.transformer._parent = self

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def user_dftype(self):
        if "dataframe_type" in self.model_fields_set:
            return self.dataframe_type
        return None

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline node is used in the context of a DLT pipeline"""
        is_orchestrator_dlt = False
        if self._parent and self._parent.is_orchestrator_dlt:
            is_orchestrator_dlt = True
        return is_orchestrator_dlt

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        if self.source is None:
            return False
        else:
            return self.source.is_cdc

    @property
    def warning_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "WARN":
                expectations[e.name] = e.expression
        return expectations

    @property
    def drop_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "DROP":
                expectations[e.name] = e.expression
        return expectations

    @property
    def fail_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "FAIL":
                expectations[e.name] = e.expression
        return expectations

    @property
    def apply_changes_kwargs(self) -> dict[str, str]:
        """Keyword arguments for dlt.apply_changes function"""
        cdc = self.source.cdc
        return {
            "apply_as_deletes": cdc.apply_as_deletes,
            "apply_as_truncates": cdc.apply_as_truncates,
            "column_list": cdc.columns,
            "except_column_list": cdc.except_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.sequence_by,
            "source": self.source.table_name,
            "stored_as_scd_type": cdc.scd_type,
            "target": self.sink.table_name,
            "track_history_column_list": cdc.track_history_columns,
            "track_history_except_column_list": cdc.track_history_except_columns,
        }

    @property
    def output_df(self) -> AnyDataFrame:
        """Node Dataframe after reading source and applying transformer."""
        return self._output_df

    @property
    def layer_spark_chain(self):
        nodes = []

        if self.layer == "BRONZE":
            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    ),
                ]

        elif self.layer == "SILVER":
            if self.timestamp_key:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_tstamp",
                            "type": "timestamp",
                            "sql_expr": self.timestamp_key,
                        },
                    )
                ]

            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_silver_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    )
                ]

        elif self.layer == "GOLD":
            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_gold_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    )
                ]

        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates
            elif self.primary_key:
                subset = [self.primary_key]

            nodes += [SparkChainNode(func_name="dropDuplicates", func_args=[subset])]

        if self.drop_source_columns and self.transformer:
            nodes += [
                SparkChainNode(
                    func_name="drop",
                    func_args=[
                        c
                        for c in self._source_columns
                        if c not in ["_bronze_at", "_silver_at", "_gold_at"]
                    ],
                )
            ]

        if len(nodes) == 0:
            return None

        return SparkChain(nodes=nodes)

    @property
    def layer_polars_chain(self):
        nodes = []

        if self.layer == "BRONZE":
            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    ),
                ]

        elif self.layer == "SILVER":
            if self.timestamp_key:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_tstamp",
                            "type": "timestamp",
                            "sql_expr": self.timestamp_key,
                        },
                    )
                ]

            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_silver_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    )
                ]

        elif self.layer == "GOLD":
            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_gold_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    )
                ]

        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates
            elif self.primary_key:
                subset = [self.primary_key]

            nodes += [PolarsChainNode(func_name="unique", func_args=[subset])]

        if self.drop_source_columns and self.transformer:
            nodes += [
                PolarsChainNode(
                    func_name="drop",
                    func_args=[
                        c
                        for c in self._source_columns
                        if c not in ["_bronze_at", "_silver_at", "_gold_at"]
                    ],
                )
            ]

        if len(nodes) == 0:
            return None

        return PolarsChain(nodes=nodes)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def get_sources(self, cls=BaseDataSource) -> list[BaseDataSource]:
        """Get all sources feeding the pipeline node"""
        sources = []

        if isinstance(self.source, cls):
            sources += [self.source]

        if self.transformer:
            sources += self.transformer.get_sources(cls)

        return sources

    def execute(
        self,
        apply_transformer: bool = True,
        spark: SparkSession = None,
        udfs: list[Callable] = None,
        write_sink: bool = True,
        full_refresh: bool = False,
    ) -> AnyDataFrame:
        """
        Execute pipeline node by:

        - Reading the source
        - Applying the user defined (and layer-specific if applicable) transformations
        - Writing the sink

        Parameters
        ----------
        apply_transformer:
            Flag to apply transformer in the execution
        spark: SparkSession
            Spark session
        udfs:
            User-defined functions
        write_sink:
            Flag to include writing sink in the execution
        full_refresh:
            If `True` dataframe will be completely re-processed by deleting
            existing data and checkpoint before processing.

        Returns
        -------
        :
            output Spark DataFrame
        """
        logger.info(f"Executing pipeline node {self.name} ({self.layer})")

        # Parse DLT
        if self.is_orchestrator_dlt:
            logger.info("DLT orchestrator selected. Sinks writing will be skipped.")
            write_sink = False
            full_refresh = False

        # Refresh
        if full_refresh:
            if self.sink:
                self.sink.purge(spark=spark)

        # Read Source
        self._output_df = self.source.read(spark)

        # Save source
        self._source_columns = self._output_df.columns

        if self.source.is_cdc and not self.is_orchestrator_dlt:
            pass
            # TODO: Apply SCD transformations
            #       Best strategy is probably to build a spark dataframe function and add a node in the chain with
            #       that function
            # https://iterationinsights.com/article/how-to-implement-slowly-changing-dimensions-scd-type-2-using-delta-table
            # https://www.linkedin.com/pulse/implementing-slowly-changing-dimension-2-using-lau-johansson-yemxf/

        # Apply transformer
        if apply_transformer:

            if "spark" in str(type(self._output_df)).lower():
                dftype = "spark"
            elif "polars" in str(type(self._output_df)).lower():
                dftype = "polars"
            else:
                raise ValueError("DataFrame type not supported")

            # Set Transformer
            transformer = self.transformer
            if transformer is None:
                if dftype == "spark":
                    transformer = SparkChain(nodes=[])
                elif dftype == "polars":
                    transformer = PolarsChain(nodes=[])
            else:
                transformer = transformer.model_copy()

            # Add layer-specific chain nodes
            if dftype == "spark" and self.layer_spark_chain:
                transformer.nodes += self.layer_spark_chain.nodes
            elif dftype == "polars" and self.layer_polars_chain:
                transformer.nodes += self.layer_polars_chain.nodes

            if transformer.nodes:
                self._output_df = transformer.execute(self._output_df, udfs=udfs)

        # Output to sink
        if write_sink and self.sink:
            self.sink.write(self._output_df)

        return self._output_df



import os
import pandas as pd
import polars as pl

from laktory.models.datasources import FileDataSource
from laktory.models.datasources import MemoryDataSource
from laktory.models.datasources import TableDataSource
from laktory._testing import Paths
from laktory._testing import spark

paths = Paths(__file__)

# DataFrame
pdf = pd.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
    },
)
df0 = spark.createDataFrame(pdf)


def test_file_data_source():
    source = FileDataSource(
        path="Volumes/sources/landing/events/yahoo_finance/stock_price"
    )

    assert source.path == "Volumes/sources/landing/events/yahoo_finance/stock_price"
    assert source.dataframe_type == "SPARK"
    assert not source.as_stream
    assert not source.is_cdc


def test_file_data_source_read():
    source = FileDataSource(
        path=os.path.join(paths.data, "./events/yahoo-finance/stock_price"),
        as_stream=False,
    )
    df = source.read(spark)
    assert df.count() == 80
    assert df.columns == [
        "data",
        "description",
        "name",
        "producer",
    ]


def test_file_data_source_polars():
    source = FileDataSource(
        path=os.path.join(
            paths.data,
            "./brz_stock_prices/part-00000-877096dd-1964-482e-9873-76361150a331-c000.snappy.parquet",
        ),
        format="PARQUET",
        dataframe_type="POLARS",
        filter="data.open > 300",
        selects={
            "data.created_at": "created_at",
            "data.symbol": "symbol",
            "data.open": "open",
            "data.close": "close",
            "data.high": "high2",
            "data.low": "low2",
            "data._created_at": "_created_at",
        },
        drops=[
            "_created_at",
        ],
        renames={
            "low2": "low",
            "high2": "high",
        },
        sample={
            "fraction": 0.5,
        },
    )
    df = source.read()

    assert df["open"].min() > 300
    assert df.columns == ["created_at", "symbol", "open", "close", "high", "low"]
    assert df.height == 10


def test_memory_data_source(df0=df0):
    source = MemoryDataSource(
        df=df0,
        filter="b != 0",
        selects=["a", "b", "c"],
        renames={"a": "aa", "b": "bb", "c": "cc"},
        broadcast=True,
    )

    # Test reader
    df = source.read(spark)
    assert df.columns == ["aa", "bb", "cc"]
    assert df.count() == 2
    # assert df.toPandas()["chain"].tolist() == ["chain", "chain"]

    # Select with rename
    source = MemoryDataSource(
        df=df0,
        selects={"x": "x1", "n": "n1"},
    )
    df = source.read(spark)
    assert df.columns == ["x1", "n1"]
    assert df.count() == 3

    # Drop
    source = MemoryDataSource(
        df=df0,
        drops=["b", "c", "n"],
    )
    df = source.read(spark)
    df.show()
    assert df.columns == ["x", "a"]
    assert df.count() == 3


def test_table_data_source():
    source = TableDataSource(
        catalog_name="dev",
        schema_name="finance",
        table_name="slv_stock_prices",
    )

    # Test meta
    assert source.full_name == "dev.finance.slv_stock_prices"
    assert source._id == "dev.finance.slv_stock_prices"


if __name__ == "__main__":
    test_file_data_source()
    test_file_data_source_read()
    test_file_data_source_polars()
    test_memory_data_source()
    test_table_data_source()



import polars as pl
import numpy as np
import polars.datatypes as T

from laktory import models

df = pl.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
        "pi": [np.pi] * 3,
        "p": [2.0, 0.2, 0.05],
        "word": ["dog_cat", "dog_cat_mouse", "dog"],
    },
)
df0 = df.select(df.columns)
df1 = df.select(df.columns)


def test_func_arg(df0=df0):
    df = df0.select(df0.columns)

    new_cols = []
    # Values to parse
    for i, v in enumerate(
        [
            "lit(3)",
            "lit(3.0)",
            "lit('3')",
            "col('x')",
            "col('x') + lit(3)",
            "sql_expr('2*x+a')",
        ]
    ):
        new_col = f"c{i}"
        a = models.PolarsChainNodeFuncArg(value=v)
        df = df.with_columns(**{new_col: a.eval()})
        new_cols += [new_col]

    # Test new column types
    data = dict(df.select(new_cols).schema)
    assert data == {
        "c0": T.Int32,
        "c1": T.Float64,
        "c2": T.String,
        "c3": T.Int64,
        "c4": T.Int64,
        "c5": T.Int64,
    }

    # Values not to parse
    for v0 in [
        "x",
        "3",
    ]:
        v1 = models.PolarsChainNodeFuncArg(value=v0).eval()
        assert v1 == v0


def test_df_input(df0=df0):
    df = df0.select(df0.columns)

    # Define Chain
    sc = models.PolarsChain(
        nodes=[
            {
                "func_name": "laktory.union",
                "func_args": [df1],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert df.height == df0.height * 2
    assert df.columns == df0.columns


def test_sql_expression(df0=df0):
    df = df0.select(df0.columns)

    sc = models.PolarsChain(
        nodes=[
            {
                "sql_expr": "SELECT *, x*2 AS x2 FROM {df}",
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    df = df.collect()
    assert df.columns == ["x", "a", "b", "c", "n", "pi", "p", "word", "x2"]
    assert df["x2"].to_list() == (df["x"] * 2).to_list()


def test_sql_with_nodes():

    sc = models.PolarsChain(
        nodes=[
            {
                "sql_expr": "SELECT * FROM {df}",
            },
            {
                "sql_expr": "SELECT * FROM {df} UNION SELECT * FROM {nodes.node_01} UNION SELECT * FROM {nodes.node_02}",
            },
        ]
    )

    assert sc.nodes[0].parsed_sql_expr.node_data_sources == []
    assert sc.nodes[1].parsed_sql_expr.node_data_sources == [
        models.PipelineNodeDataSource(node_name="node_01", dataframe_type="POLARS"),
        models.PipelineNodeDataSource(node_name="node_02", dataframe_type="POLARS"),
    ]


def test_table_input(df0=df0):
    df = df0.select(df0.columns)

    sc = models.PolarsChain(
        nodes=[
            {
                "func_name": "laktory.union",
                "func_args": [
                    {"mock_df": df},
                ],
            }
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert df.height == df0.height * 2
    assert df.columns == df0.columns


def test_column(df0=df0):
    df = df0.select(df0.columns)
    df = df.with_columns(xs=pl.lit([1, 2, 2]))

    sc = models.PolarsChain(
        nodes=[
            {
                "with_column": {
                    "name": "cos_x",
                    "type": "double",
                    "expr": "col('x').cos()",
                },
            },
            {
                "with_column": {
                    "name": "x2",
                    "type": "double",
                    "sql_expr": "x*2",
                },
            },
            {
                "with_column": {
                    "name": "xplode",
                    "type": None,
                    "expr": "pl.col('xs').list.unique()",
                },
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert [(c.name, c.expr) for c in sc.nodes[0]._with_columns] == [
        ("cos_x", "col('x').cos()")
    ]
    assert [(c.name, c.sql_expr) for c in sc.nodes[1]._with_columns] == [("x2", "x*2")]
    assert df["cos_x"].to_list() == np.cos(df["x"]).to_list()
    assert df["x2"].to_list() == (df["x"] * 2).to_list()

    # Test explode
    assert df["xplode"].to_list() == [[1, 2], [1, 2], [1, 2]]

    # Multi-columns
    sc = models.PolarsChain(
        nodes=[
            {
                "with_columns": [
                    {
                        "name": "sin_x",
                        "type": "double",
                        "expr": "col('x').sin()",
                    },
                    {
                        "name": "x3",
                        "type": "double",
                        "sql_expr": "x*3",
                    },
                ]
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert df["sin_x"].to_list() == np.sin(df["x"]).to_list()
    assert df["x3"].to_list() == (df["x"] * 3).to_list()


def test_udfs(df0=df0):
    df = df0.select(df0.columns)

    def mul3(c):
        return 3 * c

    def add_new_col(df, column_name, s=1):
        return df.with_columns(**{column_name: pl.col("x") * s})

    sc = models.PolarsChain(
        nodes=[
            {
                "with_column": {
                    "name": "rp",
                    "type": "double",
                    "expr": "pl.col('p').laktory.roundp(p=0.1)",
                },
            },
            {
                "with_column": {
                    "name": "x3",
                    "type": "double",
                    "expr": "mul3(col('x'))",
                },
            },
            {
                "func_name": "add_new_col",
                "func_args": ["y"],
                "func_kwargs": {"s": 5},
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df, udfs=[mul3, add_new_col])

    # Test
    assert df["rp"].to_list() == [2.0, 0.2, 0.1]
    assert df["x3"].to_list() == (df["x"] * 3).to_list()
    assert df["y"].to_list() == (df["x"] * 5).to_list()


def test_nested(df0=df0):
    df = df0.select(df0.columns)
    df = df.with_columns(_x2=pl.col("x").sqrt())

    sc = models.PolarsChain(
        nodes=[
            {
                "with_column": {
                    "name": "cos_x",
                    "type": "double",
                    "expr": "pl.col('x').cos()",
                },
            },
            {
                "nodes": [
                    {
                        "func_name": "rename",
                        "func_args": [
                            {"x": "x_tmp"},
                        ],
                    },
                    {
                        "with_column": {
                            "name": "x2",
                            "type": "double",
                            "expr": "pl.col('x_tmp').sqrt()",
                        },
                    },
                ],
            },
            {
                "func_name": "drop",
                "func_args": [
                    "x_tmp",
                ],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert "x_tmp" not in df.columns
    assert df["x2"].to_list() == df["_x2"].to_list()
    assert sc.columns == [
        ["x", "a", "b", "c", "n", "pi", "p", "word", "_x2"],
        ["x", "a", "b", "c", "n", "pi", "p", "word", "_x2", "cos_x"],
        ["x_tmp", "a", "b", "c", "n", "pi", "p", "word", "_x2", "cos_x", "x2"],
    ]


# TODO: Re-enable when coalesce is ready
def atest_exceptions():

    return

    df = df0.select(df0.columns)

    # Input missing - missing allowed
    sc = models.PolarsChain(
        nodes=[
            {
                "with_column": {
                    "name": "xy",
                    "type": "double",
                    "expr": "pl.Expr.coalesce('x', 'y')",
                },
            },
        ]
    )
    df = sc.execute(df)
    assert "xy" in df.columns
    #
    # # All inputs missing
    # sc = models.PolarsChain(
    #     nodes=[
    #         {
    #             "column": {"name": "xy", "type": "double"},
    #             "func_name": "coalesce",
    #             "func_args": ["col('z')", "col('y')"],
    #             "allow_missing_column_args": True,
    #         },
    #     ]
    # )
    # with pytest.raises(MissingColumnsError):
    #     df = sc.execute(df)


if __name__ == "__main__":
    test_func_arg()
    test_df_input()
    test_sql_expression()
    test_sql_with_nodes()
    test_table_input()
    test_column()
    test_udfs()
    test_nested()
    # atest_exceptions()



from __future__ import annotations
from typing import Any
from typing import Union
from typing import Literal
from typing import TYPE_CHECKING
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.constants import DEFAULT_DFTYPE
from laktory.models.basemodel import BaseModel
from laktory.models.transformers.basechainnode import BaseChainNode
from laktory.types import AnyDataFrame

if TYPE_CHECKING:
    from laktory.models.datasources.basedatasource import BaseDataSource

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class BaseChain(BaseModel):
    dataframe_type: Literal["SPARK", "POLARS"] = DEFAULT_DFTYPE
    nodes: list[Union[BaseChainNode, "BaseChain"]]
    _columns: list[list[str]] = []
    _parent: "PipelineNode" = None

    @model_validator(mode="after")
    def update_children(self) -> Any:
        for n in self.nodes:
            n._parent = self
        return self

    @model_validator(mode="after")
    def push_dftype(self) -> Any:
        dftype = self.user_dftype
        if dftype:
            for s in self.get_sources():
                s.dataframe_type = s.user_dftype or dftype
        return self

    @property
    def user_dftype(self):
        if "dataframe_type" in self.model_fields_set:
            return self.dataframe_type
        return None

    @property
    def columns(self):
        return self._columns

    def get_sources(self, cls=None) -> list[BaseDataSource]:
        """Get all sources feeding the Polars Chain"""

        from laktory.models.datasources.basedatasource import BaseDataSource

        if cls is None:
            cls = BaseDataSource

        sources = []
        for node in self.nodes:
            sources += node.get_sources(cls)
        return sources

    def execute(self, df, udfs=None) -> AnyDataFrame:
        logger.info(f"Executing {self.dataframe_type} chain")

        for inode, node in enumerate(self.nodes):
            self._columns += [df.columns]

            tnode = type(node)
            logger.info(
                f"Executing {self.dataframe_type} chain node {inode} ({tnode.__name__})."
            )
            df = node.execute(df, udfs=udfs)

        return df


BaseModel.model_rebuild()



from __future__ import annotations
import os
import json
from typing import Union
from typing import Literal
from typing import TYPE_CHECKING
from typing import Any
from pydantic import model_validator
from pydantic import Field
import networkx as nx

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.constants import CACHE_ROOT
from laktory.constants import DEFAULT_DFTYPE
from laktory.models.basemodel import BaseModel
from laktory.models.datasources.pipelinenodedatasource import PipelineNodeDataSource
from laktory.models.datasinks.tabledatasink import TableDataSink
from laktory.models.pipelinenode import PipelineNode
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.dltpipeline import DLTPipeline
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.job import JobTask
from laktory.models.resources.databricks.job import JobParameter
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.databricks.workspacefile import WorkspaceFile
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

if TYPE_CHECKING:
    from plotly.graph_objs import Figure

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class PipelineUDF(BaseModel):
    """
    Pipeline User Define Function

    Attributes
    ----------
    module_name:
        Name of the module from which the function needs to be imported.
    function_name:
        Name of the function.
    module_path:
        Workspace filepath of the module, if not in the same directory as the pipeline notebook
    """

    module_name: str
    function_name: str
    module_path: str = None


class PipelineDatabricksJob(Job):
    """
    Databricks job specifically designed to run a Laktory pipeline

    Attributes
    ----------
    laktory_version:
        Laktory version to use in the notebook tasks
    notebook_path:
        Path for the notebook. If `None`, default path for laktory job notebooks is used.

    """

    laktory_version: Union[str, None] = None
    notebook_path: Union[str, None] = None

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return super().pulumi_excludes + ["laktory_version", "notebook_path"]


class PipelineWorkspaceFile(WorkspaceFile):
    """
    Workspace File with default value for path and access controls and forced
    value for source given a pipeline name.
    """

    pipeline_name: str

    @model_validator(mode="before")
    @classmethod
    def default_values(cls, data: Any) -> Any:
        pl_name = data.get("pipeline_name", None)
        data["source"] = os.path.join(CACHE_ROOT, f"tmp-{pl_name}.json")
        if "path" not in data:
            data["path"] = f"{settings.workspace_laktory_root}pipelines/{pl_name}.json"
        if "access_controls" not in data:
            data["access_controls"] = [
                {"permission_level": "CAN_READ", "group_name": "users"}
            ]
        return data

    @property
    def resource_type_id(self):
        return "workspace-file"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return super().pulumi_excludes + ["pipeline_name"]


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class Pipeline(BaseModel, PulumiResource, TerraformResource):
    """
    Pipeline model to manage a full-fledged data pipeline including reading
    from data sources, applying data transformations through Spark and
    outputting to data sinks.

    A pipeline is composed of collections of `nodes`, each one defining its
    own source, transformations and optional sink. A node may be the source of
    another node.

    A pipeline may be run manually by using python or the CLI, but it may also
    be deployed and scheduled using one of the supported orchestrators, such as
    a Databricks Delta Live Tables or job.

    Attributes
    ----------
    databricks_job:
        Defines the Databricks Job specifications when DATABRICKS_JOB is
        selected as the orchestrator.
    dlt:
        Defines the Delta Live Tables specifications when DLT is selected as
        the orchestrator.
    name:
        Name of the pipeline
    nodes:
        List of pipeline nodes. Each node defines a data source, a series
        of transformations and optionally a sink.
    orchestrator:
        Orchestrator used for scheduling and executing the pipeline. The
        selected option defines which resources are to be deployed.
        Supported options are:
        - `DLT`: When orchestrated through Databricks DLT, each pipeline node
          creates a DLT table (or view, if no sink is defined). Behind the
          scenes, `PipelineNodeDataSource` leverages native `dlt` `read` and
          `read_stream` functions to defined the interdependencies between the
          tables as in a standard DLT pipeline. This is the recommended
          orchestrator as it is the most feature rich.
        - `DATABRICKS_JOB`: When deployed through a Databricks Job, a task
          is created for each pipeline node and all the required dependencies
          are set automatically. If a given task (or pipeline node) uses a
          `PipelineNodeDataSource` as the source, the data will be read from
          the upstream node sink.
    udfs:
        List of user defined functions provided to the transformer.
    workspacefile:
        Workspace file used to store the JSON definition of the pipeline.


    Examples
    --------
    This first example shows how to configure a simple pipeline with 2 nodes.
    Upon execution, raw data will be read from a CSV file and two DataFrames
    (bronze and silver) will be created and saved as parquet files. Notice how
    the first node is used as a data source for the second node.

    ```py
    import io
    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            format: CSV
            path: ./raw/brz_stock_prices.csv
          sink:
            format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/brz_stock_prices

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
          sink:
            format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/slv_stock_prices
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                sql_expr: data.created_at
            - with_column:
                name: symbol
                sql_expr: data.symbol
            - with_column:
                name: close
                type: double
                sql_expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
    '''

    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))

    # Execute pipeline
    # pl.execute()
    ```

    The next example defines a 3 nodes pipeline (1 bronze and 2 silvers)
    orchestrated with a Databricks Job. Notice how nodes are used as data
    sources not only for other nodes, but also for the `other` keyword argument
    of the smart join function (slv_stock_prices). Because we are using the
    DATABRICKS_JOB orchestrator, the job configuration must be declared.
    The tasks will be automatically created by the Pipeline model. Each
    task will execute a single node using the notebook referenced in
    `databricks_job.notebook_path` the content of this notebook should be
    similar to laktory.resources.notebooks.job_laktory.pl

    ```py
    import io
    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        orchestrator: DATABRICKS_JOB
        databricks_job:
          name: job-pl-stock-prices
          laktory_version: 0.3.0
          notebook_path: /Workspace/.laktory/jobs/job_laktory_pl.py
          clusters:
            - name: node-cluster
              spark_version: 14.0.x-scala2.12
              node_type_id: Standard_DS3_v2

        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/
          sink:
              path: /Volumes/dev/sources/landing/tables/dev_stock_prices/
              mode: OVERWRITE

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
          sink:
              path: /Volumes/dev/sources/landing/tables/slv_stock_prices/
              mode: OVERWRITE
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                sql_expr: data.created_at
            - with_column:
                name: symbol
                sql_expr: data.symbol
            - with_column:
                name: close
                type: double
                sql_expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
            - func_name: laktory.smart_join
              func_kwargs:
                'on':
                  - symbol
                other:
                  node_name: slv_stock_meta

        - name: slv_stock_meta
          layer: SILVER
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_meta/
          sink:
            path: /Volumes/dev/sources/landing/tables/slv_stock_meta/
            mode: OVERWRITE

    '''
    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```

    Finally, we re-implement the previous pipeline, but with a few key
    differences:

    - Orchestrator is `DLT` instead of a `DATABRICKS_JOB`
    - Sinks are Unity Catalog tables instead of storage locations
    - Data is read as a stream in most nodes
    - `slv_stock_meta` is simply a DLT view since it does not have an associated
      sink.

    We also need to provide some basic configuration for the DLT pipeline.

    ```py
    import io
    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        orchestrator: DLT
        dlt:
            catalog: dev
            target: sandbox
            access_controls:
            - group_name: users
              permission_level: CAN_VIEW

        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/
            as_stream: true
          sink:
            table_name: brz_stock_prices

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
            as_stream: true
          sink:
            table_name: slv_stock_prices
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                sql_expr: data.created_at
            - with_column:
                name: symbol
                sql_expr: data.symbol
            - with_column:
                name: close
                type: double
                sql_expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
            - func_name: laktory.smart_join
              func_kwargs:
                'on':
                  - symbol
                other:
                  node_name: slv_stock_meta

        - name: slv_stock_meta
          layer: SILVER
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_meta/

    '''
    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```

    References
    ----------
    * [Databricks Job](https://docs.databricks.com/en/workflows/jobs/create-run-jobs.html)
    * [Databricks DLT](https://www.databricks.com/product/delta-live-tables)

    """

    databricks_job: Union[PipelineDatabricksJob, None] = None
    dataframe_type: Literal["SPARK", "POLARS"] = DEFAULT_DFTYPE
    dlt: Union[DLTPipeline, None] = None
    name: str
    nodes: list[Union[PipelineNode]]
    orchestrator: Union[Literal["DLT", "DATABRICKS_JOB"], None] = None
    udfs: list[PipelineUDF] = []
    workspacefile: PipelineWorkspaceFile = None

    @model_validator(mode="before")
    @classmethod
    def assign_name_to_dlt(cls, data: Any) -> Any:

        if "dlt" in data.keys():
            data["dlt"]["name"] = data.get("name", None)

        return data

    @model_validator(mode="before")
    @classmethod
    def assign_name_to_workspacefile(cls, data: Any) -> Any:
        workspacefile = data.get("workspacefile", None)
        if workspacefile:
            workspacefile["pipeline_name"] = data["name"]
        return data

    @model_validator(mode="before")
    @classmethod
    def push_dftype_before(cls, data: Any) -> Any:
        dftype = data.get("dataframe_type", None)
        if dftype:
            if "nodes" in data.keys():
                for n in data["nodes"]:
                    if isinstance(n, dict):
                        n["dataframe_type"] = n.get("dataframe_type", dftype)
        return data

    @model_validator(mode="after")
    def push_dftype_after(self) -> Any:
        dftype = self.user_dftype
        if dftype:
            for node in self.nodes:
                node.dataframe_type = node.user_dftype or dftype
                node.push_dftype_after()
        return self

    @model_validator(mode="after")
    def update_children(self) -> Any:
        # Build dag
        _ = self.dag

        # Assign pipeline
        for n in self.nodes:
            n._parent = self

        return self

    @model_validator(mode="after")
    def validate_dlt(self) -> Any:

        if not self.is_orchestrator_dlt:
            return self

        if self.dlt is None:
            raise ValueError("dlt must be defined if DLT orchestrator is selected.")

        for n in self.nodes:

            # Sink must be Table or None
            if n.sink is None:
                pass
            if isinstance(n.sink, TableDataSink):
                if n.sink.table_name != n.name:
                    raise ValueError(
                        "For DLT pipeline, table sink name must be the same as node id."
                    )
                n.sink.catalog_name = self.dlt.catalog
                n.sink.schema_name = self.dlt.target

        return self

    @model_validator(mode="after")
    def validate_job(self) -> Any:

        if not self.orchestrator == "DATABRICKS_JOB":
            return self

        if self.databricks_job is None:
            raise ValueError(
                "databricks_job must be defined if DATABRICKS_JOB orchestrator is selected."
            )

        cluster_found = False
        for c in self.databricks_job.clusters:
            if c.name == "node-cluster":
                cluster_found = True
        # if not cluster_found:
        #     raise ValueError(
        #         "To use DATABRICKS_JOB orchestrator, a cluster named `node-cluster` must be defined in the databricks_job attribute."
        #     )

        job = self.databricks_job
        job.parameters = [
            JobParameter(name="pipeline_name", default=self.name),
            JobParameter(name="full_refresh", default="false"),
        ]

        notebook_path = self.databricks_job.notebook_path
        if notebook_path is None:
            notebook_path = f"{settings.workspace_laktory_root}jobs/job_laktory_pl.py"

        package = "laktory"
        if self.databricks_job.laktory_version:
            package += f"=={self.databricks_job.laktory_version}"

        job.tasks = []

        # Sorting Node Names to prevent job update trigger with Pulumi
        node_names = [node.name for node in self.nodes]
        node_names.sort()
        for node_name in node_names:

            node = self.nodes_dict[node_name]

            depends_on = []
            for edge in self.dag.in_edges(node.name):
                depends_on += [{"task_key": "node-" + edge[0]}]

            job.tasks += [
                JobTask(
                    task_key="node-" + node.name,
                    notebook_task={
                        "base_parameters": {"node_name": node.name},
                        "notebook_path": notebook_path,
                    },
                    # libraries=[{"pypi": {"package": package}}],
                    depends_ons=depends_on,
                    # job_cluster_key=job_cluster_key,
                )
            ]
            if cluster_found:
                job.tasks[-1].job_cluster_key = "node-cluster"

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def user_dftype(self):
        if "dataframe_type" in self.model_fields_set:
            return self.dataframe_type
        return None

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline orchestrator is DLT"""
        return self.orchestrator == "DLT"

    @property
    def nodes_dict(self) -> dict[str, PipelineNode]:
        """
        Nodes dictionary whose keys are the node names.

        Returns
        -------
        :
            Nodes
        """
        return {n.name: n for n in self.nodes}

    @property
    def dag(self) -> nx.DiGraph:
        """
        Networkx Directed Acyclic Graph representation of the pipeline. Useful
        to identify interdependencies between nodes.

        Returns
        -------
        :
            Directed Acyclic Graph
        """

        # TODO: Review if this needs to be computed dynamically or if we can
        #       compute it only after initialization. It depends if we believe
        #       that the nodes will be changed after initialization.

        dag = nx.DiGraph()

        # Build nodes
        for n in self.nodes:
            dag.add_node(n.name)

        # Build edges and assign nodes to pipeline node data sources
        for n in self.nodes:
            for s in n.get_sources(PipelineNodeDataSource):
                dag.add_edge(s.node_name, n.name)
                if s.node_name not in self.nodes_dict:
                    raise ValueError(
                        f"Pipeline node data source '{s.node_name}' is not defined in pipeline '{self.name}'"
                    )
                s.node = self.nodes_dict[s.node_name]

        if not nx.is_directed_acyclic_graph(dag):
            raise ValueError(
                "Pipeline is not a DAG (directed acyclic graph)."
                " A circular dependency has been detected. Please review nodes dependencies."
            )

        return dag

    @property
    def sorted_node_names(self):
        return list(nx.topological_sort(self.dag))

    @property
    def sorted_nodes(self) -> list[PipelineNode]:
        """
        Topologically sorted nodes.

        Returns
        -------
        :
            List of Topologically sorted nodes.
        """
        node_names = self.sorted_node_names
        nodes = []
        for node_name in node_names:
            for node in self.nodes:
                if node.name == node_name:
                    nodes += [node]
                    break

        return nodes

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def execute(
        self, spark=None, udfs=None, write_sinks=True, full_refresh: bool = False
    ) -> None:
        """
        Execute the pipeline (read sources and write sinks) by sequentially
        executing each node. The selected orchestrator might impact how
        data sources or sinks are processed.

        Parameters
        ----------
        spark:
            Spark Session
        udfs:
            List of user-defined functions used in transformation chains.
        write_sinks:
            If `False` writing of node sinks will be skipped
        full_refresh:
            If `True` all nodes will be completely re-processed by deleting
            existing data and checkpoints before processing.
        """
        logger.info("Executing Pipeline")

        for inode, node in enumerate(self.sorted_nodes):
            node.execute(
                spark=spark,
                udfs=udfs,
                write_sink=write_sinks,
                full_refresh=full_refresh,
            )

    def dag_figure(self) -> Figure:
        """
        [UNDER DEVELOPMENT] Generate a figure representation of the pipeline
        DAG.

        Returns
        -------
        :
            Plotly figure representation of the pipeline.
        """
        import plotly.graph_objs as go

        dag = self.dag
        pos = nx.spring_layout(dag)

        # ------------------------------------------------------------------- #
        # Edges                                                               #
        # ------------------------------------------------------------------- #

        edge_traces = []
        for e in dag.edges:
            edge_traces += [
                go.Scatter(
                    x=[pos[e[0]][0], pos[e[1]][0]],
                    y=[pos[e[0]][1], pos[e[1]][1]],
                    line={
                        "color": "#a006b1",
                    },
                    marker=dict(
                        symbol="arrow-bar-up",
                        angleref="previous",
                        size=30,
                    ),
                    mode="lines+markers",
                    hoverinfo="none",
                    showlegend=False,
                )
            ]

        # ------------------------------------------------------------------- #
        # Nodes                                                               #
        # ------------------------------------------------------------------- #

        nodes_trace = go.Scatter(
            x=[_p[0] for _p in pos.values()],
            y=[_p[1] for _p in pos.values()],
            text=list(dag.nodes),
            name="pl nodes",
            mode="markers+text",
            marker=dict(
                size=50,
                color="#06d49e",
                line=dict(
                    width=2,
                    color="#dff2ed",
                ),
            ),
            textfont=dict(
                color="#a006b1",
            ),
        )

        return go.Figure(data=[nodes_trace] + edge_traces)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def self_as_core_resources(self):
        return False

    @property
    def resource_type_id(self) -> str:
        """
        pl
        """
        return "pl"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - configuration workspace file
        - configuration workspace file permissions

        if orchestrator is `DLT`:

        - DLT Pipeline

        if orchestrator is `DATABRICKS_JOB`:

        - Databricks Job
        """
        # Configuration file
        source = os.path.join(CACHE_ROOT, f"tmp-{self.name}.json")
        d = self.model_dump(exclude_unset=True)
        d = self.inject_vars(d)
        s = json.dumps(d, indent=4)
        with open(source, "w", newline="\n") as fp:
            fp.write(s)

        resources = []
        if self.orchestrator in ["DATABRICKS_JOB", "DLT"]:

            file = self.workspacefile
            if file is None:
                file = PipelineWorkspaceFile(pipeline_name=self.name)
            resources += [file]

            if self.is_orchestrator_dlt:
                resources += [self.dlt]

            if self.orchestrator == "DATABRICKS_JOB":
                resources += [self.databricks_job]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return ""  # "databricks:Pipeline"

    #
    # @property
    # def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return {
    #         "access_controls": True,
    #         "tables": True,
    #         "clusters": {"__all__": {"access_controls"}},
    #         "udfs": True,
    #     }
    #
    # @property
    # def pulumi_properties(self):
    #     d = super().pulumi_properties
    #     k = "clusters"
    #     if k in d:
    #         _clusters = []
    #         for c in d[k]:
    #             c["label"] = c.pop("name")
    #             _clusters += [c]
    #         d[k] = _clusters
    #     return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return ""

    #
    # @property
    # def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return self.pulumi_excludes
    #
    # @property
    # def terraform_properties(self) -> dict:
    #     d = super().terraform_properties
    #     k = "cluster"
    #     if k in d:
    #         _clusters = []
    #         for c in d[k]:
    #             c["label"] = c.pop("name")
    #             _clusters += [c]
    #         d[k] = _clusters
    #     return d



from typing import Any
from pydantic import model_validator

from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.types import AnyDataFrame
from laktory.spark import SparkDataFrame
from laktory.spark import is_spark_dataframe
from laktory.polars import PolarsDataFrame
from laktory.polars import is_polars_dataframe
from laktory._logger import get_logger

logger = get_logger(__name__)


class MemoryDataSource(BaseDataSource):
    """
    Data source using in-memory DataFrame, generally used in the context of a
    data pipeline.

    Attributes
    ----------
    df
        Input DataFrame

    Examples
    ---------
    ```python
    from laktory import models
    import pandas as pd

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    source = models.MemoryDataSource(
        df=df,
        as_stream=False,
    )
    df = source.read()
    ```
    """

    df: Any

    @model_validator(mode="after")
    def set_dataframe_type(self) -> Any:

        if is_spark_dataframe(self.df):
            dataframe_type = "SPARK"
        elif is_polars_dataframe(self.df):
            dataframe_type = "POLARS"
        else:
            raise ValueError("DataFrame must be of type Spark or Polars")

        self.dataframe_type = dataframe_type

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #
    #
    # @property
    # def _id(self):
    #     return str(self.path)

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:
        logger.info(f"Reading {self._id} from memory")
        return self.df

    def _read_polars(self) -> PolarsDataFrame:
        logger.info(f"Reading {self._id} from memory")
        return self.df



import os
from typing import Any
from typing import Literal
from typing import Union
from pydantic import model_validator
from pydantic import Field
from laktory import constants
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions


class NotebookLookup(ResourceLookup):
    """
    Attributes
    ----------
    path:
        Notebook path on the workspace
    format:
        Notebook format to export. Either `SOURCE`, `HTML`, `JUPYTER`, or `DBC`
    """

    path: str = Field(serialization_alias="id")
    format: str = "SOURCE"


class Notebook(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Notebook

    Attributes
    ----------
    access_controls:
        List of notebook access controls
    dirpath:
        Workspace directory containing the notebook. Filename will be assumed to be the same as local filepath. Used
        if path is not specified.
    language:
         Notebook programming language
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    path:
        Workspace filepath for the notebook
    source:
        Path to notebook in source code format on local filesystem.

    Examples
    --------
    ```py
    from laktory import models

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(notebook.path)
    #> /.laktory/dlt/dlt_laktory_pl.py

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/create_view.py", dirpath="/views/"
    )
    print(notebook.path)
    #> /views/create_view.py
    ```
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    language: Literal["SCALA", "PYTHON", "SQL", "R"] = None
    lookup_existing: NotebookLookup = Field(None, exclude=True)
    path: str = None
    source: str

    @property
    def filename(self) -> str:
        """Notebook file name"""
        return os.path.basename(self.source)

    @model_validator(mode="after")
    def default_path(self) -> Any:
        if self.path is None:
            if self.dirpath:
                self.path = f"{self.dirpath}{self.filename}"

            elif "/notebooks/" in self.source:
                self.path = (
                    constants.LAKTORY_WORKSPACE_ROOT
                    + self.source.split("/notebooks/")[-1]
                )

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        key = self.path
        if key is None:
            return ""
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    notebook_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Notebook"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_notebook"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from laktory.models.resources.databricks import VectorSearchEndpoint

vector_search_endpoint = VectorSearchEndpoint(
    name="default",
)


def test_vector_search_endpoint():
    print(vector_search_endpoint)
    assert vector_search_endpoint.name == "default"


if __name__ == "__main__":
    test_vector_search_endpoint()



import os
import shutil
import networkx as nx
import polars

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths

paths = Paths(__file__)

OPEN_FIGURES = False


with open(os.path.join(paths.data, "pl-spark-local.yaml"), "r") as fp:
    pl = models.Pipeline.model_validate_yaml(fp)


with open(os.path.join(paths.data, "pl-polars-local.yaml"), "r") as fp:
    pl_polars = models.Pipeline.model_validate_yaml(fp)

with open(os.path.join(paths.data, "pl-polars-local.yaml"), "r") as fp:
    pl_polars2 = models.Pipeline.model_validate_yaml(fp)

with open(os.path.join(paths.data, "pl-spark-dlt.yaml"), "r") as fp:
    pl_dlt = models.Pipeline.model_validate_yaml(fp)  # also used in test_stack


# Update paths
brz_sink_path = os.path.join(paths.tmp, "pl_brz_sink")
slv_sink_path = os.path.join(paths.tmp, "pl_slv_sink")
gld_sink_path = os.path.join(paths.tmp, "pl_gld_sink")
meta_sink_path = os.path.join(paths.tmp, "pl_slv_meta_sink")
for _pl in [pl, pl_polars, pl_polars2]:
    _pl.nodes[0].source.path = os.path.join(paths.data, "brz_stock_prices")
    _pl.nodes[3].source.path = os.path.join(paths.data, "slv_stock_meta")
    _pl.nodes[0].sink.path = os.path.join(paths.tmp, brz_sink_path)
    _pl.nodes[1].sink.path = os.path.join(paths.tmp, slv_sink_path)
    _pl.nodes[2].sink.path = os.path.join(paths.tmp, gld_sink_path)
    _pl.nodes[3].sink.path = os.path.join(paths.tmp, meta_sink_path)
    if _pl.name == pl_polars.name:
        for i in [0, 3]:
            _dirpath = _pl.nodes[i].source.path
            filenames = [
                filename
                for filename in os.listdir(_dirpath)
                if filename.endswith(".parquet")
            ]
            _pl.nodes[i].source.path = os.path.join(_dirpath, filenames[0])

# Save nodes
node_brz = pl.nodes[0]
node_slv = pl.nodes[1]
node_gld = pl.nodes[2]
node_meta = pl.nodes[3]


def test_dag():
    dag = pl.dag

    # Test Dag
    assert nx.is_directed_acyclic_graph(dag)
    assert len(dag.nodes) == 4
    assert len(dag.edges) == 3
    assert list(nx.topological_sort(dag)) == [
        "brz_stock_prices",
        "slv_stock_meta",
        "slv_stock_prices",
        "gld_max_stock_prices",
    ]

    # Test nodes assignment
    assert pl.sorted_nodes == [node_brz, node_meta, node_slv, node_gld]
    assert node_slv.source.node == node_brz
    assert node_gld.source.node == node_slv
    assert (
        node_slv.transformer.nodes[-1].parsed_func_kwargs["other"].value.node
        == node_meta
    )

    # Test figure
    fig = pl.dag_figure()
    fig.write_html(os.path.join(paths.tmp, "dag.html"), auto_open=OPEN_FIGURES)


def test_children():

    for pn in pl.nodes:
        assert pn._parent == pl
        assert pn.source._parent == pn
        if pn.transformer is not None:
            assert pn.transformer._parent == pn
        if pn.sink is not None:
            assert pn.sink._parent == pn

        if pn.transformer:
            for tn in pn.transformer.nodes:
                assert tn._parent == pn.transformer

        for s in pn.get_sources():
            assert s._parent == pn


def test_execute():

    pl.execute(spark)

    # In memory DataFrames
    assert pl.nodes_dict["brz_stock_prices"].output_df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "_bronze_at",
    ]
    assert pl.nodes_dict["brz_stock_prices"].output_df.count() == 80
    assert pl.nodes_dict["slv_stock_meta"].output_df.columns == [
        "symbol2",
        "currency",
        "first_traded",
        "_silver_at",
    ]
    assert pl.nodes_dict["slv_stock_meta"].output_df.count() == 3
    assert pl.nodes_dict["slv_stock_prices"].output_df.columns == [
        "_bronze_at",
        "created_at",
        "close",
        "currency",
        "first_traded",
        "_silver_at",
        "symbol",
    ]
    assert pl.nodes_dict["slv_stock_prices"].output_df.count() == 80
    assert pl.nodes_dict["gld_max_stock_prices"].output_df.columns == [
        "symbol",
        "max_price",
        "min_price",
        "_gold_at",
    ]
    assert pl.nodes_dict["gld_max_stock_prices"].output_df.count() == 4

    # Sinks
    _df_slv = spark.read.format("PARQUET").load(slv_sink_path)
    _df_gld = spark.read.format("PARQUET").load(gld_sink_path)
    assert _df_slv.columns == [
        "_bronze_at",
        "created_at",
        "close",
        "currency",
        "first_traded",
        "_silver_at",
        "symbol",
    ]
    assert _df_slv.count() == 80
    assert _df_gld.columns == ["symbol", "max_price", "min_price", "_gold_at"]
    assert _df_gld.count() == 4

    # Cleanup
    shutil.rmtree(brz_sink_path)
    shutil.rmtree(slv_sink_path)
    shutil.rmtree(gld_sink_path)
    shutil.rmtree(meta_sink_path)


def test_execute_node():

    # Execute each node individually and clear intermediate output dataframe
    for inode, node in enumerate(pl.sorted_nodes):
        node.execute(spark=spark)
        if inode < len(pl.nodes) - 1:
            node._output_df = None

    assert pl.nodes_dict["gld_max_stock_prices"].output_df.count() == 4

    # Cleanup
    shutil.rmtree(brz_sink_path)
    shutil.rmtree(slv_sink_path)
    shutil.rmtree(gld_sink_path)
    shutil.rmtree(meta_sink_path)


def test_pipeline_sql():
    _pl = pl.model_copy()

    # Select join node
    node = _pl.nodes_dict["slv_stock_prices"]
    t4 = node.transformer.nodes[4]

    t4.sql_expr = """
    SELECT 
        *
    FROM 
        {df} as df
    LEFT JOIN 
        {nodes.slv_stock_meta} as meta
    ON df.symbol = meta.symbol2
    ;
    """

    _pl.execute(spark=spark, write_sinks=False)
    df = node.output_df
    assert df.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "symbol2",
        "currency",
        "first_traded",
        "_silver_at",
    ]


def test_execute_polars():

    _pl = pl_polars

    # Check dataframe type assignment
    assert _pl.dataframe_type == "POLARS"
    for node in _pl.nodes:
        assert node.dataframe_type == "POLARS"
        assert node.source.dataframe_type == "POLARS"
        for s in node.get_sources():
            assert s.dataframe_type == "POLARS"

    _pl.execute()

    # In memory DataFrames
    assert _pl.nodes_dict["brz_stock_prices"].output_df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "_bronze_at",
    ]
    assert _pl.nodes_dict["brz_stock_prices"].output_df.height == 80
    assert _pl.nodes_dict["slv_stock_meta"].output_df.columns == [
        "symbol2",
        "currency",
        "first_traded",
        "_silver_at",
    ]
    assert _pl.nodes_dict["slv_stock_meta"].output_df.height == 3
    assert _pl.nodes_dict["slv_stock_prices"].output_df.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "currency",
        "first_traded",
        "_silver_at",
    ]
    assert _pl.nodes_dict["slv_stock_prices"].output_df.height == 80
    assert _pl.nodes_dict["gld_max_stock_prices"].output_df.columns == [
        "symbol",
        "max_price",
        "min_price",
        "_gold_at",
    ]
    assert _pl.nodes_dict["gld_max_stock_prices"].output_df.height == 4

    # Sinks
    _df_slv = polars.read_parquet(slv_sink_path)
    _df_gld = polars.read_parquet(gld_sink_path)
    assert _df_slv.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "currency",
        "first_traded",
        "_silver_at",
    ]
    assert _df_slv.height == 80
    assert _df_gld.columns == ["symbol", "max_price", "min_price", "_gold_at"]
    assert _df_gld.height == 4

    # Cleanup
    os.remove(brz_sink_path)
    os.remove(slv_sink_path)
    os.remove(gld_sink_path)
    os.remove(meta_sink_path)


def test_execute_polars_sql():

    _pl = pl_polars2.model_copy()

    # Select join node
    node = _pl.nodes_dict["slv_stock_prices"]
    t3 = node.transformer.nodes[3]

    t3.sql_expr = """
    SELECT 
        *
    FROM 
        {df} as df
    LEFT JOIN 
        {nodes.slv_stock_meta} as meta
    ON df.symbol = meta.symbol2
    ;
    """

    _pl.execute(spark=spark, write_sinks=False)
    df = node.output_df
    assert df.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "symbol2",
        "currency",
        "first_traded",
        "_silver_at",
    ]


def test_pipeline_dlt():

    # Test Sink as Source
    sink_source = pl_dlt.nodes[1].source.node.sink.as_source(
        as_stream=pl_dlt.nodes[1].source.as_stream
    )
    data = sink_source.model_dump()
    print(data)
    assert data == {
        "as_stream": True,
        "broadcast": False,
        "cdc": None,
        "dataframe_type": "SPARK",
        "drops": None,
        "filter": None,
        "limit": None,
        "renames": None,
        "sample": None,
        "selects": None,
        "watermark": None,
        "catalog_name": "dev",
        "table_name": "brz_stock_prices",
        "schema_name": "sandbox",
        "warehouse": "DATABRICKS",
    }

    data = pl_dlt.model_dump()
    print(data)
    assert data == {
        "databricks_job": None,
        "dataframe_type": "SPARK",
        "dlt": {
            "access_controls": [
                {
                    "group_name": "account users",
                    "permission_level": "CAN_VIEW",
                    "service_principal_name": None,
                    "user_name": None,
                }
            ],
            "allow_duplicate_names": None,
            "catalog": "dev",
            "channel": "PREVIEW",
            "clusters": [],
            "configuration": {},
            "continuous": None,
            "development": None,
            "edition": None,
            "libraries": None,
            "name": "pl-stock-prices",
            "notifications": [],
            "photon": None,
            "serverless": None,
            "storage": None,
            "target": "sandbox",
        },
        "name": "pl-stock-prices",
        "nodes": [
            {
                "add_layer_columns": True,
                "dlt_template": "DEFAULT",
                "dataframe_type": "SPARK",
                "description": None,
                "drop_duplicates": None,
                "drop_source_columns": False,
                "transformer": None,
                "expectations": [],
                "layer": "BRONZE",
                "name": "brz_stock_prices",
                "primary_key": None,
                "sink": {
                    "mode": None,
                    "write_options": {},
                    "catalog_name": "dev",
                    "checkpoint_location": None,
                    "format": "DELTA",
                    "schema_name": "sandbox",
                    "table_name": "brz_stock_prices",
                    "warehouse": "DATABRICKS",
                },
                "source": {
                    "as_stream": True,
                    "broadcast": False,
                    "cdc": None,
                    "dataframe_type": "SPARK",
                    "drops": None,
                    "filter": None,
                    "limit": None,
                    "renames": None,
                    "sample": None,
                    "selects": None,
                    "watermark": None,
                    "format": "JSON",
                    "header": True,
                    "multiline": False,
                    "path": "/Volumes/dev/sources/landing/events/yahoo-finance/stock_price/",
                    "read_options": {},
                    "schema_location": None,
                },
                "timestamp_key": None,
            },
            {
                "add_layer_columns": True,
                "dlt_template": "DEFAULT",
                "dataframe_type": "SPARK",
                "description": None,
                "drop_duplicates": None,
                "drop_source_columns": True,
                "transformer": {
                    "dataframe_type": "SPARK",
                    "nodes": [
                        {
                            "dataframe_type": "SPARK",
                            "func_args": [],
                            "func_kwargs": {},
                            "func_name": None,
                            "sql_expr": None,
                            "with_column": {
                                "name": "created_at",
                                "type": "timestamp",
                                "unit": None,
                                "expr": None,
                                "sql_expr": "data.created_at",
                            },
                            "with_columns": [],
                        },
                        {
                            "dataframe_type": "SPARK",
                            "func_args": [],
                            "func_kwargs": {},
                            "func_name": None,
                            "sql_expr": None,
                            "with_column": {
                                "name": "symbol",
                                "type": "string",
                                "unit": None,
                                "expr": None,
                                "sql_expr": "data.symbol",
                            },
                            "with_columns": [],
                        },
                        {
                            "dataframe_type": "SPARK",
                            "func_args": [],
                            "func_kwargs": {},
                            "func_name": None,
                            "sql_expr": None,
                            "with_column": {
                                "name": "close",
                                "type": "double",
                                "unit": None,
                                "expr": None,
                                "sql_expr": "data.close",
                            },
                            "with_columns": [],
                        },
                        {
                            "dataframe_type": "SPARK",
                            "func_args": ["data", "producer", "name", "description"],
                            "func_kwargs": {},
                            "func_name": "drop",
                            "sql_expr": None,
                            "with_column": None,
                            "with_columns": [],
                        },
                    ],
                },
                "expectations": [],
                "layer": "SILVER",
                "name": "slv_stock_prices",
                "primary_key": None,
                "sink": {
                    "mode": None,
                    "write_options": {},
                    "catalog_name": "dev",
                    "checkpoint_location": None,
                    "format": "DELTA",
                    "schema_name": "sandbox",
                    "table_name": "slv_stock_prices",
                    "warehouse": "DATABRICKS",
                },
                "source": {
                    "as_stream": True,
                    "broadcast": False,
                    "cdc": None,
                    "dataframe_type": "SPARK",
                    "drops": None,
                    "filter": None,
                    "limit": None,
                    "renames": None,
                    "sample": None,
                    "selects": None,
                    "watermark": None,
                    "node_name": "brz_stock_prices",
                },
                "timestamp_key": None,
            },
        ],
        "orchestrator": "DLT",
        "udfs": [],
        "workspacefile": None,
    }

    # Test resources
    resources = pl_dlt.core_resources
    assert len(resources) == 4

    wsf = resources[0]
    wsfp = resources[1]
    dlt = resources[2]
    dltp = resources[3]

    assert isinstance(wsf, models.resources.databricks.WorkspaceFile)
    assert isinstance(wsfp, models.resources.databricks.Permissions)
    assert isinstance(dlt, models.resources.databricks.DLTPipeline)
    assert isinstance(dltp, models.resources.databricks.Permissions)

    assert dlt.resource_name == "dlt-pl-stock-prices"
    assert dltp.resource_name == "permissions-dlt-pl-stock-prices"
    assert wsf.resource_name == "workspace-file-laktory-pipelines-pl-stock-prices-json"
    assert (
        wsfp.resource_name
        == "permissions-workspace-file-laktory-pipelines-pl-stock-prices-json"
    )

    assert dlt.options.provider == "${resources.databricks2}"
    assert dltp.options.provider == "${resources.databricks2}"
    assert wsf.options.provider == "${resources.databricks1}"
    assert wsfp.options.provider == "${resources.databricks1}"

    assert dlt.options.depends_on == []
    assert dltp.options.depends_on == ["${resources.dlt-pl-stock-prices}"]
    assert wsf.options.depends_on == []
    assert wsfp.options.depends_on == [
        "${resources.workspace-file-laktory-pipelines-pl-stock-prices-json}"
    ]


def test_pipeline_job():

    with open(os.path.join(paths.data, "pl-spark-job.yaml"), "r") as fp:
        pl = models.Pipeline.model_validate_yaml(fp)

    # Test job
    job = pl.databricks_job
    data = job.model_dump(exclude_unset=True)
    print(data)
    assert data == {
        "clusters": [
            {
                "name": "node-cluster",
                "node_type_id": "Standard_DS3_v2",
                "spark_version": "14.0.x-scala2.12",
            }
        ],
        "name": "job-pl-stock-prices",
        "parameters": [
            {"default": "pl-stock-prices", "name": "pipeline_name"},
            {"default": "false", "name": "full_refresh"},
        ],
        "tasks": [
            {
                "depends_ons": [],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "brz_stock_prices"},
                },
                "task_key": "node-brz_stock_prices",
            },
            {
                "depends_ons": [],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "slv_stock_meta"},
                },
                "task_key": "node-slv_stock_meta",
            },
            {
                "depends_ons": [
                    {"task_key": "node-brz_stock_prices"},
                    {"task_key": "node-slv_stock_meta"},
                ],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "slv_stock_prices"},
                },
                "task_key": "node-slv_stock_prices",
            },
        ],
        "laktory_version": "0.3.0",
    }

    # Test resources
    resources = pl.core_resources
    assert len(resources) == 3


if __name__ == "__main__":
    test_dag()
    test_children()
    test_execute()
    test_execute_node()
    test_pipeline_sql()
    test_execute_polars()
    test_execute_polars_sql()
    test_pipeline_dlt()
    test_pipeline_job()



from laktory.constants import SUPPORTED_DATATYPES


def test_constants():
    assert "double" in SUPPORTED_DATATYPES
    assert "str" not in SUPPORTED_DATATYPES


if __name__ == "__main__":
    test_constants()



from laktory.models.resources.databricks import VectorSearchIndex

vector_search_index = VectorSearchIndex(
    name="dev.finances.market_news_vs_index",
    primary_key="id",
    endpoint_name="default",
    index_type="DELTA_SYNC",
    delta_sync_index_spec={
        "source_table": "dev.finances.market_news",
        "embedding_vector_columns": [
            {"name": "embedding", "embedding_dimension": 4095}
        ],
        "pipeline_type": "TRIGGERED",
    },
)


def test_vector_search_index():
    print(vector_search_index)
    assert vector_search_index.name == "dev.finances.market_news_vs_index"
    assert vector_search_index.delta_sync_index_spec.pipeline_type == "TRIGGERED"


if __name__ == "__main__":
    test_vector_search_index()



import os
from typing import Any
from typing import Union
from pydantic import model_validator
from pydantic import Field
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions


class DbfsFileLookup(ResourceLookup):
    """
    Attributes
    ----------
    path:
        Path on DBFS for the file from which to get content.
    limit_file_size:
        Do not load content for files larger than 4MB.
    """

    path: str = Field(serialization_alias="id")
    limit_file_size: bool = True


class DbfsFile(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks DBFS File

    Attributes
    ----------
    access_controls:
        List of file access controls
    dirpath:
        Workspace directory containing the file. Filename will be assumed to be the same as local filepath. Used if path
        is not specified.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    path:
         DBFS filepath for the file
    source:
        Path to file on local filesystem.
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    lookup_existing: DbfsFileLookup = Field(None, exclude=True)
    path: str = None
    source: str

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {"path": ""}

    @property
    def filename(self) -> str:
        """File filename"""
        return os.path.basename(self.source)

    @model_validator(mode="after")
    def default_path(self) -> Any:
        if self.path is None:
            if self.dirpath:
                self.path = f"{self.dirpath}{self.filename}"
            else:
                raise ValueError(
                    "A value for `dirpath` must be specified if `path` is not specified"
                )

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        # key = os.path.splitext(self.path)[0]
        key = self.path
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    workspace_file_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:DbfsFile"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_dbfs_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes
